package com.cch_lab.android.apply.prototype.tinybasic;

import com.cch_lab.android.apply.prototype.tinybasic.command.CommandManager;
import com.cch_lab.android.apply.prototype.tinybasic.command.CommandRuntimeEnvironment;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxTreeStream;
import com.cch_lab.android.apply.prototype.tinybasic.source.SourceStream;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.AbstractSyntax;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxHolder;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxManager;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxPatterns;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxStream;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxTree;
import com.cch_lab.android.apply.prototype.tinybasic.token.AbstractToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.TokenManager;
import com.cch_lab.android.apply.prototype.tinybasic.token.TokenSource;
import com.cch_lab.android.apply.prototype.tinybasic.token.TokenStream;
import java.nio.file.Files;
import java.io.BufferedReader;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Paths;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;


/**
 * コンパイラ・クラス.
 * <p>
 *     <ul>
 *         スクリプト・ソースからコンパイルまでの以下の処理の実行を担います。
 *         <li>字句解析</li>
 *         <li>構文解析</li>
 *         <li><strike>抽象構文木最適化</strike></li>
 *         <li>命令解析</li>
 *         <li>命令実行</li>
 *     </ul>
 * </p>
 */
public class Compiler {
    /** デバッグフラグ. */
    public static boolean DEBUG = false;

    /**
     * メイン・メソッド.
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        Compiler compiler = new Compiler();
        
        if (args.length == 1) {
            compiler.run(args[0]);
        }
//        compiler.test();
    }

    public void test() {
        List<String> sourceLines = new ArrayList<>();
        sourceLines.add("String name = \"John\";");
        sourceLines.add("String greet = \"Hello\";");
        sourceLines.add("print(greet + \" \" + name + \"!\");");
        sourceLines.add("Integer num = +1+(2+3)*4+5-6;");
        sourceLines.add("print(num);");

        // デバッグ出力 ON
        this.DEBUG = true;
        
        // ソースコード生成
        SourceStream sourceStream = new SourceStream(sourceLines);


        // 字句解析
        TokenStream tokenStream = executeLexicalParse(sourceStream, new TokenManager());

        // TODO デバッグ出力を消すこと
        debug("トークン展開 (定義順)");
        for (int index = 0; index < tokenStream.size(); index++) {
            AbstractToken token = tokenStream.getToken(index);
            debug(token.getClass().getSimpleName() + ":" + token.getTokenSource().getSource());
        }


        // 構文解析
        SyntaxTree syntaxTree = executeSyntaxParse(tokenStream, new SyntaxManager());

        // TODO デバッグ出力を消すこと
        debug("トークン展開 (定義順)");
        List<AbstractToken> tokens = SyntaxTreeStream.extractTokenStream(syntaxTree);
        for (AbstractToken token : tokens) {
            debug("Token["+(tokens.indexOf(token))+"]="
                    + token.getClass().getSimpleName()
                    +"\n  '" + token.getTokenSource().getSource() + "'");
        }

        // TODO デバッグ出力を消すこと
        debug("構文木展開 (定義順)");
        List<AbstractSyntax> syntaxTrees = SyntaxTreeStream.extractSyntaxStream(syntaxTree);
        for (AbstractSyntax syntax : syntaxTrees) {
            SyntaxHolder prev = syntax.getPrevSyntax();
            SyntaxHolder body = syntax.getBodySyntax();
            SyntaxHolder next = syntax.getNextSyntax();
            String prevInfo = debugSyntaxTree(prev, syntaxTrees);
            String bodyInfo = debugSyntaxTree(body, syntaxTrees);
            String nextInfo = debugSyntaxTree(next, syntaxTrees);
            debug("SyntaxTree["+(syntaxTrees.indexOf(syntax))+"]=" + syntax.getClass().getSimpleName());
            if (prevInfo != null) debug("  prev" + prevInfo);
            if (bodyInfo != null) debug("  body" + bodyInfo);
            if (nextInfo != null) debug("  next" + nextInfo);
        }

        // TODO デバッグ出力を消すこと
        debug("構文木展開 (入れ子構造順)");
        SyntaxTreeStream syntaxTreeStream = new SyntaxTreeStream(syntaxTree);
        for (int index = 0; index < syntaxTreeStream.size(); index++) {
            SyntaxTree node = syntaxTreeStream.getSyntaxTree(index);
            debug(syntaxTreeStream.getSyntaxTreeInfo(node));
        }


        // 構文最適化
        // FIXME 現時点では、最適化しません。


        // 命令解析
        CommandRuntimeEnvironment runtime = executeCommandParse(syntaxTree, new CommandManager());


        // 命令実行
        debug("命令実行 （構文木を評価完了順に実行）"); // TODO デバッグ出力を消すこと
        executeCommand(runtime);
    }

    /**
     * プログラム実行.
     * @param path ソースコード・パス
     */
    public void run(String path) {
        // デバッグ出力 OFF
        this.DEBUG = false;
        
        try {
            // ソースコード生成
            SourceStream sourceStream = getSourceStream(path);

            // 字句解析
            TokenStream tokenStream = executeLexicalParse(sourceStream, new TokenManager());

            // 構文解析
            SyntaxTree syntaxTree = executeSyntaxParse(tokenStream, new SyntaxManager());

            // 構文最適化
            // FIXME 現時点では、最適化しません。

            // 命令解析
            CommandRuntimeEnvironment runtime = executeCommandParse(syntaxTree, new CommandManager());

            // 命令実行
            executeCommand(runtime);
            
        } catch (Exception e) {
            System.err.println(e.getMessage());
        }
    }

    /**
     * ソース取得.
     * @param path ソースコード・パス
     * @return ソース・ストリーム
     */
    public SourceStream getSourceStream(String path) {
        
        List<String> sourceLines = new ArrayList<>();
        try (BufferedReader reader = Files.newBufferedReader(Paths.get(path), Charset.forName("UTF-8"));) {
            while (true) {
                String sourceLine = reader.readLine();
                if (sourceLine != null){
                    sourceLines.add(sourceLine);
                } else {
                    break;
                }
            }
        } catch(IOException e) {
            throw new RuntimeException(e);
        }
        
        return new SourceStream(sourceLines);
    }
    
    /**
     * 字句解析実行.
     * @param sourceStream ソースコード・ストリーム
     * @param tokenManager トークン・マネジャ (利用可能なトークンと優先度を管理します)
     * @return トークン・ストリーム
     */
    public TokenStream executeLexicalParse(SourceStream sourceStream, TokenManager tokenManager){
        TokenStream tokenStream = new TokenStream();
        sourceStream.rewind();
        AbstractToken prevToken = null;

        while (sourceStream.hasNextScanSource()) {
            tokenManager.rewind();
            boolean hasToken = false;
            // TODO デバッグ出力を消すこと
            if (Compiler.DEBUG)
                debug("check source = " + sourceStream.getScanSource());

            while (tokenManager.hasNext()) {
                try {
                    AbstractToken token = tokenManager.getToken();
                    if (token.isPattern(prevToken, sourceStream)) {
                        tokenStream.add(token);
                        hasToken = true;
                        // TODO デバッグ出力を消すこと
                        if (Compiler.DEBUG)
                            debug("check token = " + token.getClass().getSimpleName()
                                +" is match. source='"+token.getTokenSource().getSource()+"'");

                        // 次のスキャン・ソースコード位置に進む。
                        sourceStream.nextScanSource(token.getTokenSource().getSource());
                        break;
                    }
                } catch (IllegalAccessException e) {
                    System.out.println(e.getMessage());
                    e.printStackTrace();
                } catch (InstantiationException e) {
                    System.out.println(e.getMessage());
                    e.printStackTrace();
                }
            }

            if (!hasToken) {
                throw new IllegalStateException(
                        createErrorMessage(
                                "Illegal Statement Error:",
                                sourceStream.getSourceLine(sourceStream.getScanRow()),
                                sourceStream.getScanRow(),
                                sourceStream.getScanColmn()));
            }
        }
        return tokenStream;
    }

    /**
     * 構文解析実行.
     * @param tokenStream   トークン・ストリーム
     * @param syntaxManager 構文マネジャ (利用可能な構文と優先度を管理します)
     * @return 構文木 (構文木ルート)
     */
    public SyntaxTree executeSyntaxParse(TokenStream tokenStream, SyntaxManager syntaxManager) {
        SyntaxStream syntaxStream = new SyntaxStream(tokenStream);

        boolean hasSyntax = true;
        try {
            syntaxManager.rewind();
            while (syntaxManager.hasNext()) {
                // チェック対象構文取得
                AbstractSyntax checkerSyntax = null;
                checkerSyntax = syntaxManager.getSyntax();

                // 構文木解析
                int parseCount;
                do {
                    parseCount = 0;
                    syntaxStream.rewind();
                    while (syntaxStream.hasNext()) {
                        boolean isMatch = checkerSyntax.isPattern(syntaxStream);
                        if (isMatch){
                            // TODO デバッグ出力を消すこと
                            if (Compiler.DEBUG)
                                debugSyntax(checkerSyntax, syntaxStream);

                            // チェック対象を入れ子にして、次の構文に進む
                            syntaxStream.nextSyntax(true, checkerSyntax);
                            checkerSyntax = syntaxManager.getSyntax();
                            parseCount++;

                            // TODO デバッグ出力を消すこと
                            if (Compiler.DEBUG)
                                debug("  after index="+syntaxStream.getScanIndex()+", syntaxStream.size()="+syntaxStream.size());
                        } else{
                            // 次の構文に進む
                            syntaxStream.nextSyntax(false, checkerSyntax);
                        }
                    }
                } while (parseCount > 0) ;
            }
        } catch (IllegalAccessException e) {
            System.out.println(e.getMessage());
            e.printStackTrace();
            hasSyntax = false;
        } catch (InstantiationException e) {
            System.out.println(e.getMessage());
            e.printStackTrace();
            hasSyntax = false;
        }
        if (!hasSyntax) {
            throw new IllegalStateException(
                    createErrorMessage(
                            "Illegal Statement Error:" +
                                    "the syntax '" +
                                    syntaxManager.getSyntaxClass().getSimpleName() +
                                    "' can not create."
                    )
            );
        }

        // ルート構文 (収束構文) 取得
        AbstractSyntax rootSyntax = null;
        try {
            rootSyntax = syntaxManager.getSyntax(syntaxManager.getSize()-1);
        } catch (InstantiationException e) {
            throw new RuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }

        if (syntaxStream.size() == 1 &&
                rootSyntax.getClass().equals(syntaxStream.getSyntax(0).getTypeClass())) {
            // TODO デバッグ出力を消すこと
            if (Compiler.DEBUG)
                debug("\n構文解析成功\n");
        }else{
            // 構文木解析による入れ子化が、収束しなかった場合
            System.err.println("構文解析失敗 error syntax count=" + syntaxStream.size());
            LinkedHashMap<String, String> messageMap = new LinkedHashMap<>();
            for (int index = 0; index < syntaxStream.size(); index++) {
                SyntaxHolder syntax = syntaxStream.getSyntax(index);
                if (rootSyntax.getClass().getSimpleName().equals(
                        syntax.getTypeClass().getSimpleName())) {
                    // 最終段階まで解析されているのでチェックしない。
                    continue;
                }

                // エラー推定箇所出力
                Range range = getErrorRange(syntax);
                {
                    StringBuilder sb = new StringBuilder();
                    /*
                    String minPosition = createErrorMessage(
                            "",
                            range.sourceStream.getSourceLine(range.minRow),
                            range.minRow);
                    */
                    String minPosition = createErrorMessage(
                            "",
                            range.sourceStream.getSourceLine(range.minRow),
                            range.minRow,
                            range.minStart);
                    sb.append("Syntax error:");
                    sb.append(minPosition).append("\n");
                    messageMap.put(sb.toString(), sb.toString());
                }
            }
            StringBuilder sb = new StringBuilder();
            for (String message : messageMap.values()) {
                sb.append(message);
            }
            throw new IllegalStateException(
                    createErrorMessage(
                            "Illegal Statement Error:failed to syntax parse.\n" + sb.toString())
            );
        }

        // ルート構文を格納する、構文木を生成する
        final SyntaxHolder root = syntaxStream.getSyntax(0);
        AbstractSyntax top = new AbstractSyntax() {
            @Override
            public SyntaxHolder getBodySyntax() {
                return root;
            }
            @Override
            public SyntaxPatterns getSyntaxPatterns() {
                return null;
            }
        };
        return new SyntaxTree(new SyntaxHolder(top));
    }

    /**
     * コマンド解析実行.
     * @param syntaxTree  構文木 (ルート構文木)
     * @param commandManager コマンド・マネジャ (利用可能な命令(組込機能)と優先度を管理します)
     * @return コマンド・ランタイム実行環境
     */
    public CommandRuntimeEnvironment executeCommandParse(SyntaxTree syntaxTree, CommandManager commandManager){
        return new CommandRuntimeEnvironment(syntaxTree, commandManager);
    }

    /**
     * コマンド実行.
     * @param runtime コマンド・ランタイム実行環境
     * @return 実行結果
     */
    public boolean executeCommand(CommandRuntimeEnvironment runtime){
        return runtime.evaluate(0) != null;
    }


    private String createErrorMessage(String message, String sourceLine, int sourceRow) {
        StringBuilder sb = new StringBuilder();
        sb.append("\n");
        sb.append(message).append(" at near line=").append(sourceRow).append("\n");
        sb.append(sourceLine).append("\n");
        return sb.toString();
    }

    private String createErrorMessage(String message, String sourceLine, int sourceRow, int sourceColumn) {
        final String SPACE_512 =
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                ";

        StringBuilder sb = new StringBuilder();
        sb.append("\n");
        sb.append(message).append(" at near line=").append(sourceRow).
                append(", column=").append(sourceColumn).append(")\n");
        sb.append(sourceLine).append("\n");
        sb.append(SPACE_512.substring(0, sourceColumn)).append("^").append("\n");
        return sb.toString();
    }

    private String createErrorMessage(String message) {
        StringBuilder sb = new StringBuilder();
        sb.append("\n");
        sb.append(message).append("\n");
        return sb.toString();
    }

    // 構文木展開デバッグ出力
    private String debugSyntaxTree(SyntaxHolder syntax, List<AbstractSyntax> stream) {
        return syntax == null
                ? null
                : syntax.getSyntax() != null
                    ? "["+(stream.indexOf(syntax.getSyntax()))+"]="+syntax.getSyntax().getClass().getSimpleName()
                    : syntax.getToken() != null
                        ? "="+syntax.getToken().getTokenSource().getSource()
                        : null;
    }

    // 汎用デバッグ出力
    private void debug(String message) {
        System.out.println(message);
    }

    // 構文デバッグ出力
    private void debugSyntax(AbstractSyntax checkerSyntax, SyntaxStream syntaxStream) {
        SyntaxHolder prev = checkerSyntax.getPrevSyntax();
        SyntaxHolder body = checkerSyntax.getBodySyntax();
        SyntaxHolder next = checkerSyntax.getNextSyntax();
        System.out.println("match syntax=" + checkerSyntax.getClass().getSimpleName());
        TokenSource tokenSource = getTokenSource(body);
        System.out.println("  sourceLine["+tokenSource.getRow()+"]="+tokenSource.getSourceStream().getSourceLine(tokenSource.getRow()));
        System.out.println("  before index="+syntaxStream.getScanIndex()+", syntaxStream.size()="+syntaxStream.size());
        if (prev != null && prev.getType() == SyntaxHolder.Type.TOKEN) {
            System.out.println("  prev=" + prev.getTypeClass().getSimpleName()
                    + ", '"+prev.getToken().getTokenSource().getSource()+"'");
        }
        if (prev != null && prev.getType() == SyntaxHolder.Type.SYNTAX) {
            System.out.println("  prev=" + prev.getTypeClass().getSimpleName());
        }
        if (body != null && body.getType() == SyntaxHolder.Type.TOKEN) {
            System.out.println("  body=" + body.getTypeClass().getSimpleName()
                    + ", '"+body.getToken().getTokenSource().getSource()+"'");
        }
        if (body != null && body.getType() == SyntaxHolder.Type.SYNTAX) {
            System.out.println("  body=" + body.getTypeClass().getSimpleName());
        }
        if (next != null && next.getType() == SyntaxHolder.Type.TOKEN) {
            System.out.println("  next=" + next.getTypeClass().getSimpleName()
                    + ", '"+next.getToken().getTokenSource().getSource()+"'");
        }
        if (next != null && next.getType() == SyntaxHolder.Type.SYNTAX) {
            System.out.println("  next=" + next.getTypeClass().getSimpleName());
        }
    }

    private TokenSource getTokenSource(SyntaxHolder holder) {
        do {
            if (holder.getType() == SyntaxHolder.Type.SYNTAX) {
                holder = holder.getSyntax().getBodySyntax();
            }
        } while (holder.getType() != SyntaxHolder.Type.TOKEN);
        return holder.getToken().getTokenSource();
    }

    // エラーソースの範囲を取得する
    private Range getErrorRange(SyntaxHolder holder) {
        Range range = new Range();
        range.set(Integer.MAX_VALUE, Integer.MAX_VALUE, -1, -1);
        getErrorRange(holder, range);
        return range;
    }
    private void getErrorRange(SyntaxHolder holder, Range p) {
        if (holder.getType()== SyntaxHolder.Type.SYNTAX) {
            AbstractSyntax syntax = holder.getSyntax();
            SyntaxHolder prev = syntax.getPrevSyntax();
            SyntaxHolder body = syntax.getBodySyntax();
            SyntaxHolder next = syntax.getNextSyntax();
            if (prev != null) getErrorRange(prev, p);
            if (body != null) getErrorRange(body, p);
            if (next != null) getErrorRange(next, p);
        } else
        if (holder.getType()== SyntaxHolder.Type.TOKEN) {
            AbstractToken token = holder.getToken();
            TokenSource tokenSource = token.getTokenSource();
            int row   = tokenSource.getRow();
            int start = tokenSource.getStart();
            int end   = tokenSource.getEnd();
            int minRow   = p.minRow > row ? row : p.minRow;
            int minStart = p.minRow > row ? start : p.minStart > start ? start : p.minStart;
            int maxRow   = p.maxRow < row ? row : p.maxRow;
            int maxEnd   = p.maxRow < row ? end : p.maxEnd < end ? end : p.maxEnd;
            p.set(minRow, minStart, maxRow, maxEnd - 1);
            p.set(token.getTokenSource().getSourceStream());
        }
    }
    private class Range {
        public SourceStream sourceStream;
        public int minRow;
        public int minStart;
        public int maxRow;
        public int maxEnd;
        public Range() {
        }
        public void set(SourceStream sourceStream) {
            this.sourceStream = sourceStream;
        }
        public void set(int minRow, int minStart, int maxRow, int maxEnd) {
            this.minRow   = minRow;
            this.minStart = minStart;
            this.maxRow   = maxRow;
            this.maxEnd   = maxEnd;
        }
    }

}


