package com.cch_lab.android.apply.prototype.tinybasic.command;

import com.cch_lab.android.apply.prototype.tinybasic.Compiler;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.AbstractSyntax;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxTree;


/**
 * コマンド基底(抽象)クラス.
 */
public abstract class AbstractCommand {
    private Class<? extends AbstractSyntax>[] supportSyntax;

    /** コマンド・パラメータ構文木. */
    private SyntaxTree syntaxTree;


    /**
     * コマンド・サポート構文取得.
     * @return コマンド・サポート構文
     */
    public abstract Class<? extends AbstractSyntax>[] getSupportSyntax();

    /**
     * コマンドパターン・チェック.
     * @param runtime    コマンド実行時環境
     * @param syntaxTree 構文木
     * @return チェック結果 (true:構文木はサポート対象、false:構文木はサポート対象でない)
     */
    public boolean isPattern(CommandRuntimeEnvironment runtime, SyntaxTree syntaxTree) {
        if (supportSyntax == null) {
            supportSyntax = getSupportSyntax();
        }

        boolean isPattern = false;
        if (supportSyntax != null) {
            for (Class<?> checkerSyntax : supportSyntax) {
                if (checkerSyntax.equals(syntaxTree.getNodeSyntax().getClass())) {
                    this.syntaxTree = syntaxTree;
                    isPattern = true;
                    break;
                }
            }
        }
        return isPattern;
    }

    /**
     * コマンド評価.
     * @param runtime コマンド実行時環境
     * @return 評価結果 (nullの場合もあり)
     */
    public Object evaluate(CommandRuntimeEnvironment runtime) {
        int parentIndex     = runtime.getParentIndex(syntaxTree.getNode());
        int nodeIndex       = runtime.getIndex(syntaxTree.getNode());
        int leftLeafIndex   = runtime.getIndex(syntaxTree.getLeftLeaf());   // leftLeaf  は、構文の prev に対応する
        int middleLeafIndex = runtime.getIndex(syntaxTree.getMiddleLeaf()); // middleLeafは、構文の body に対応する
        int rightLeafIndex  = runtime.getIndex(syntaxTree.getRightLeaf());  // rightLeaf は、構文の next に対応する

        Object leftLeafResult   = runtime.evaluate(leftLeafIndex);
        Object middleLeafResult = runtime.evaluate(middleLeafIndex);
        Object rightLeafResult  = runtime.evaluate(rightLeafIndex);

        // TODO デバッグ出力をオフにすること
        if (Compiler.DEBUG) {
            // 対応構文木のノード構文名をデバッグ出力する
            String targetName = getParameterSyntaxTree().getNodeSyntax().getClass().getSimpleName();
            System.out.println(this.getClass().getSimpleName() + "  target="+targetName);
        }
        return evaluate(runtime, leftLeafResult, middleLeafResult, rightLeafResult);
    }

    /**
     * コマンド評価.
     * <p>
     *     実装先では、実装コマンド用に引数評価結果をキャストして、
     *     実装コマンド用の評価結果を返してください。
     * </p>
     * @param runtime          コマンド実行時環境
     * @param leftLeafResult   左リーフ評価結果 (null の場合もあり)
     * @param middleLeafResult 中リーフ評価結果 (null の場合もあり)
     * @param rightLeafResult  右リーフ評価結果 (null の場合もあり)
     * @return 評価結果 (nullの場合もあり)
     */
    public abstract Object evaluate(
            CommandRuntimeEnvironment runtime,
            Object leftLeafResult, Object middleLeafResult, Object rightLeafResult);

    /**
     * コマンド・パラメータ構文木取得.
     * @return コマンドパラメータ構文木
     */
    public SyntaxTree getParameterSyntaxTree() {
        return syntaxTree;
    }
}
