package com.cch_lab.android.apply.prototype.tinybasic.command;

import java.util.ArrayList;
import java.util.List;


/**
 * コマンド・マネージャ基底(抽象)クラス.
 * <p>
 *     コマンドの優先順位を管理するクラスです。<br/>
 *     setup メソッドで管理対象とするコマンドを登録してください。
 *     register メソッドで登録した順番が優先順位となります。
 * </p>
 */
public abstract class AbstractCommandManager {
    private int scanIndex;
    private List<Class<? extends AbstractCommand>> commandList = new ArrayList<>();

    /**
     * コンストラクタ.
     */
    public AbstractCommandManager() {
        setup();
    }

    /**
     * セットアップ．
     * <p>
     *     コマンド追加(登録)設定を行ってください。
     *     <ol>登録順序(優先順序)基準
     *         <li>自己構文が、他のコマンド・パターンに含まれるもの(依存性が高いもの)</li>
     *         <li>優先度の高いもの順(乗除算 > 加減算)であること<li/>
     *         <li>小さな概念順(演算 < 演算式 < 文)であること</li>
     *     </ol>
     * </p>
     */
    public abstract void setup();

    /**
     * コマンド登録.
     * @param command コマンド
     */
    public void register(Class<? extends AbstractCommand> command) {
        commandList.add(command);
    }

    /**
     * 巻き戻し.
     */
    public void rewind() {
        scanIndex = 0;
    }

    /**
     * コマンド保持チェック.
     * <p>
     *     次のコマンドがあるか否かをチェックします。<br/>
     *     コマンドがある場合、getCommand()で取得できます。
     * </p>
     * @return コマンド保持有無 (true:あり、false:なし)
     */
    public boolean hasNext() {
        return commandList.size() > scanIndex;
    }

    /**
     * コマンド取得.
     * @return コマンド
     */
    public AbstractCommand getCommand() throws InstantiationException, IllegalAccessException {
        return commandList.get(scanIndex++).newInstance();
    }

    /**
     * コマンドクラス取得.
     * @return コマンドクラス
     */
    public Class<?> getCommandClass() {
        return commandList.get(scanIndex);
    }

    /**
     * コマンド個数取得.
     * @return コマンド個数
     */
    public int size() {
        return commandList.size();
    }

    /**
     * コマンド取得.
     * @param index コマンド取得位置インデックス
     * @return コマンド
     */
    public AbstractCommand getCommand(int index) throws InstantiationException, IllegalAccessException {
        return commandList.get(index).newInstance();
    }

    /**
     * コマンドクラス取得.
     * @param index コマンド取得位置インデックス
     * @return コマンのクラス
     */
    public Class<?> getCommandClass(int index) {
        return commandList.get(index);
    }

}
