package com.cch_lab.android.apply.prototype.tinybasic.command;

import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxTree;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxTreeStream;
import com.cch_lab.android.apply.prototype.tinybasic.token.AbstractToken;

/**
 * コマンド実行環境.
 * <p>
 *     コマンドの実行環境を提供するクラスです。
 * </p>
 */
public class CommandRuntimeEnvironment {

    private SyntaxTreeStream syntaxTreeStream;
    private CommandStream    commandStream;
    private VariableSpace    variableSpace;

    /**
     * コンストラクタ.
     * @param root           プログラムルート・構文木
     * @param commandManager コマンド・マネージャ
     */
    public CommandRuntimeEnvironment(SyntaxTree root, AbstractCommandManager commandManager) {
        syntaxTreeStream = new SyntaxTreeStream(root);
        commandStream    = new CommandStream(syntaxTreeStream);
        variableSpace    = new VariableSpace();

        try {
            for (int index = 0; index < syntaxTreeStream.size(); index++) {
                SyntaxTree syntaxTree = syntaxTreeStream.getSyntaxTree(index);

                // 構文木のノード構文ごとに対応するコマンドを探索/生成して、
                // コマンド列をつくり上げる。
                commandManager.rewind();
                while(commandManager.hasNext()) {
                    AbstractCommand command = commandManager.getCommand();
                    if (command.isPattern(this, syntaxTree)) {
                        commandStream.add(command);
                        break;
                    }
                }
            }
        } catch(InstantiationException e) {
            e.printStackTrace();
            throw new IllegalStateException(e);
        } catch(IllegalAccessException e) {
            e.printStackTrace();
            throw new IllegalStateException(e);
        }
    }

    /**
     * （親）コマンド位置インデックス取得.
     * <p>
     *     構文木に対応するコマンドが配置された、
     *     コマンド・ストリーム内の位置インデックスを取得します。
     * </p>
     * @param syntaxTree 構文木
     * @return （親）コマンド位置インデックス (対応先が見つからないか引数がルートの場合、-1が返ります)
     */
    public int getParentIndex(SyntaxTree syntaxTree) {
        return syntaxTreeStream.getParentIndex(syntaxTree);
    }

    /**
     * コマンド位置インデックス取得.
     * <p>
     *     構文木に対応するコマンドが配置された、
     *     コマンド・ストリーム内の位置インデックスを取得します。
     * </p>
     * @param syntaxTree 構文木
     * @return コマンド位置インデックス (対応先が見つからない場合、-1が返ります)
     */
    public int getIndex(SyntaxTree syntaxTree) {
        return syntaxTreeStream.getIndex(syntaxTree);
    }

    /**
     * コマンド取得.
     * <p>
     *     コマンド・ストリーム内のコマンドを取得します。
     * </p>
     * @param index コマンド位置インデックス
     * @return コマンド (nullの場合もあり)
     */
    public AbstractCommand getCommand(int index) {
        if (index == -1) {
            return null;
        }
        return commandStream.getCommand(index);
    }

    /**
     * コマンド評価.
     * <p>
     *     コマンド・ストリーム内のコマンドを評価します。
     * </p>
     * @param index コマンド位置インデックス
     * @return 評価結果 (nullの場合もあり)
     */
    public Object evaluate(int index) {
        if (index == -1) {
            return null;
        }
        return commandStream.getCommand(index).evaluate(this);
    }


    /**
     * 変数宣言.
     * @param name  変数名
     * @param type  変数型
     * @param token 対応トークン
     */
    public void defineVariable(String name, Class<?> type, AbstractToken token) throws CommandRuntimeException {
        variableSpace.defineVariable(name, type, token);
    }

    /**
     *
     * 変数型取得.
     * @param name  変数名
     * @param token 対応トークン
     * @return 変数型
     * @exception CommandRuntimeException 対応する変数が未定義の場合に発生します。
     */
    public Class<?> getVariableType(String name, AbstractToken token) throws CommandRuntimeException {
        return variableSpace.getValueType(name, token);
    }

    /**
     * 変数値取得.
     * @param name  変数名
     * @param type 変数型
     * @param token 対応トークン
     * @return 変数値 (取得側では、変数型に従った型キャストが必要)
     * @exception CommandRuntimeException 対応する変数が未定義か、型が異なる場合に発生します。
     */
    public Object getVariableValue(String name, Class<?> type, AbstractToken token) throws CommandRuntimeException {
        return variableSpace.getValue(name, type, token);
    }

    /**
     * 変数値取得.
     * @param name  変数名
     * @param value 変数値
     * @param token 対応トークン
     * @exception CommandRuntimeException 対応する変数が未定義か、型が異なる場合に発生します。
     */
    public void setVariableValue(String name, Object value, AbstractToken token) throws CommandRuntimeException {
        variableSpace.setValue(name, value, token);
    }

    /**
     * エラー情報得.
     * @param  token 変数トークン
     * @return エラー情報文字列
     */
    public String getErrorInfo(AbstractToken token) {
        return variableSpace.createErrorInfo(token);
    }
}
