package com.cch_lab.android.apply.prototype.tinybasic.command;

import com.cch_lab.android.apply.prototype.tinybasic.token.AbstractToken;

import java.util.HashMap;


/**
 * 変数空間・クラス.
 * <p>
 *     変数の定義や値の設定と取得を管理するクラスです。
 * </p>
 */
class VariableSpace {
    private HashMap<String, Value> variableSpace;

    /**
     * デフォルト・コンストラクタ.
     */
    VariableSpace() {
        variableSpace = new HashMap<>();
    }

    /**
     * 変数宣言.
     * @param name  変数名
     * @param type  変数型
     * @param token 対応トークン
     */
    void defineVariable(String name, Class<?> type, AbstractToken token) throws CommandRuntimeException {
        if (variableSpace.containsKey(name)) {
            throw new CommandRuntimeException(
                    "Variable name '" + name + "’ is already defined. "
                            + createErrorInfo(token));
        }
        Value variable = new Value(type);
        variableSpace.put(name, variable);
    }

    /**
     * 変数型取得.
     * @param name  変数名
     * @param token 対応トークン
     * @return 変数型
     * @exception CommandRuntimeException 対応する変数が未定義の場合に発生します。
     */
    Class<?> getValueType(String name, AbstractToken token) throws CommandRuntimeException {
        if (!variableSpace.containsKey(name)) {
            throw new CommandRuntimeException(
                    "Variable name '" + name + "’ is not defined. "
                            + createErrorInfo(token));
        }
        return variableSpace.get(name).type;
    }

    /**
     * 変数値取得.
     * @param name  変数名
     * @param type 変数型
     * @param token 対応トークン
     * @return 変数値 (取得側では、変数型に従った型キャストが必要)
     * @exception CommandRuntimeException 対応する変数が未定義の場合に発生します。
     */
    Object getValue(String name, Class<?> type, AbstractToken token) throws CommandRuntimeException {
        if (!variableSpace.containsKey(name)) {
            throw new CommandRuntimeException(
                    "Variable name '" + name + "’ is not defined. "
                            + createErrorInfo(token));
        }
        Class<?> variableType = getValueType(name, token);
        if (!variableType.equals(type)) {
            throw new CommandRuntimeException(
                    "Variable name '" + name + "’ defined type is '" + variableType.getSimpleName() + "', "
                            +"but expected type is ’" + type.getSimpleName() + "’. "
                            + createErrorInfo(token));
        }
        return variableSpace.get(name).getValue();
    }

    /**
     * 変数値取得.
     * @param name  変数名
     * @param value 変数値
     * @param token 対応トークン
     * @exception CommandRuntimeException 対応する変数が未定義か、型が異なる場合に発生します。
     */
    void setValue(String name, Object value, AbstractToken token) throws CommandRuntimeException {
        if (!variableSpace.containsKey(name)) {
            throw new CommandRuntimeException(
                    "Variable name '" + name + "’ is not defined. "
                            + createErrorInfo(token));
        }
        Value variable = variableSpace.get(name);
        if (value != null && !variable.type.equals(value.getClass())) {
            throw new CommandRuntimeException(
                    "Variable name '" + name + "’ defined type is '" + variable.type.getSimpleName() + "', "
                            +"but value type is ’" + value.getClass().getSimpleName() + "’. "
                            + createErrorInfo(token));
        }
        variable.setValue(value);
        variableSpace.put(name, variable);
    }

    /**
     * 変数エラー情報得.
     * @param  token 変数トークン
     * @return 変数エラー情報文字列
     */
    String createErrorInfo(AbstractToken token) {
        final String SPACE_512 =
                "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                "+
                        "                                                                ";

        int row     = token.getTokenSource().getRow();
        int start   = token.getTokenSource().getStart();
        String line = token.getTokenSource().getSourceStream().getSourceLine(row);

        StringBuilder sb = new StringBuilder();
        sb.append(" at (").append(row).append(",").append(start).append(")\n");
        sb.append(line).append("\n");
        sb.append(SPACE_512.substring(0, start)).append("^");
        return sb.toString();
    }
}

