package com.cch_lab.android.apply.prototype.tinybasic.command.item;

import com.cch_lab.android.apply.prototype.tinybasic.command.AbstractCommand;
import com.cch_lab.android.apply.prototype.tinybasic.command.CommandRuntimeEnvironment;
import com.cch_lab.android.apply.prototype.tinybasic.command.CommandRuntimeException;
import com.cch_lab.android.apply.prototype.tinybasic.command.Void;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.AbstractSyntax;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.SyntaxTree;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.item.PrimaryBinaryOperator;
import com.cch_lab.android.apply.prototype.tinybasic.syntax.item.SecondaryBinaryOperator;
import com.cch_lab.android.apply.prototype.tinybasic.token.AbstractToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.DivideOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.EqualsOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.GreaterEqualsOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.GreaterThenOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.LessEqualsOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.LessThenOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.MinusOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.MultiOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.NotEqualsOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.PlusOperatorToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.item.ReminderOperatorToken;


/**
 * 二項演算・コマンド.
 * <p>
 *     二項演算を表すコマンドです。
 * </p>
 */
public class BinaryOperatorCommand extends AbstractCommand {

    /**
     *{@inheritDoc}
     */
    @Override
    public Class<? extends AbstractSyntax>[] getSupportSyntax() {
        Class[] patterns =
                    new Class[]{
                            PrimaryBinaryOperator.class,
                            SecondaryBinaryOperator.class
                    };
        return (Class<? extends AbstractSyntax>[]) patterns;
    }

    /**
     *{@inheritDoc}
     */
    @Override
    public Object evaluate(
            CommandRuntimeEnvironment runtime,
            Object leftLeafResult, Object middleLeafResult, Object rightLeafResult) {

        // 対応構文ノードとトークン＆構文取得
        SyntaxTree node = getParameterSyntaxTree();
        AbstractToken token = node.getMiddleToken();

        // 二項演算を評価して返却
        return BinaryOperatorCommand.evaluate(runtime, token, leftLeafResult, middleLeafResult, rightLeafResult);
    }

    /**
     * 二項演算評価.
     * <p>
     *     汎用の二項演算評価を外部に提供します。
     * </p>
     * @param runtime          コマンド実行時環境
     * @param token            コマンド・ソースを保持するトークン
     * @param leftLeafResult   左リーフ評価結果 (null の場合もあり)
     * @param middleLeafResult 中リーフ評価結果 (null の場合もあり)
     * @param rightLeafResult  右リーフ評価結果 (null の場合もあり)
     * @return 評価結果 (nullの場合もあり)
     */
    public static Object evaluate(
            CommandRuntimeEnvironment runtime,
            AbstractToken token,
            Object leftLeafResult, Object middleLeafResult, Object rightLeafResult) {

            // 有効値チェック
            if (leftLeafResult == null) {
                throw new CommandRuntimeException(
                        "NullPointerException: " + runtime.getErrorInfo(token));
            } else
            if(leftLeafResult instanceof Void) {
                throw new CommandRuntimeException(
                        "The return value is a void: " + runtime.getErrorInfo(token));
            } else
            if (rightLeafResult == null) {
                throw new CommandRuntimeException(
                        "NullPointerException: " + runtime.getErrorInfo(token));
            } else
            if(rightLeafResult instanceof Void) {
                throw new CommandRuntimeException(
                        "The return value is a void: " + runtime.getErrorInfo(token));
            }

            // 値型取得
            Class<?> leftType  = leftLeafResult.getClass();
            Class<?> rightType = rightLeafResult.getClass();

            // 二項演算値返却
            if(token.getClass().equals(PlusOperatorToken.class)) {
                if (leftType.equals(Integer.class) && rightType.equals(Integer.class)) {
                    return ((Integer)leftLeafResult) + ((Integer)rightLeafResult);
                } else
                if (leftType.equals(String.class) && rightType.equals(String.class)) {
                    return ((String)leftLeafResult) + ((String)rightLeafResult);
                }
            } else
            if(token.getClass().equals(MinusOperatorToken.class)) {
                if (leftType.equals(Integer.class) && rightType.equals(Integer.class)) {
                    return ((Integer)leftLeafResult) - ((Integer)rightLeafResult);
                }
            } else
            if(token.getClass().equals(MultiOperatorToken.class)) {
                if (leftType.equals(Integer.class) && rightType.equals(Integer.class)) {
                    return ((Integer)leftLeafResult) * ((Integer)rightLeafResult);
                }
            } else
            if(token.getClass().equals(DivideOperatorToken.class)) {
                if (leftType.equals(Integer.class) && rightType.equals(Integer.class)) {
                    return ((Integer)leftLeafResult) / ((Integer)rightLeafResult);
                }
            } else
            if(token.getClass().equals(ReminderOperatorToken.class)) {
                if (leftType.equals(Integer.class) && rightType.equals(Integer.class)) {
                    return ((Integer)leftLeafResult) % ((Integer)rightLeafResult);
                }
            } else
            if (leftType.equals(Integer.class) && rightType.equals(Integer.class)) {
                // 整数の論理演算
                if(token.getClass().equals(EqualsOperatorToken.class)) {
                    return (leftLeafResult).equals(rightLeafResult);
                } else
                if(token.getClass().equals(NotEqualsOperatorToken.class)) {
                    return !(leftLeafResult).equals(rightLeafResult);
                } else
                if(token.getClass().equals(GreaterThenOperatorToken.class)) {
                    return ((Integer)leftLeafResult) > ((Integer)rightLeafResult);
                } else
                if(token.getClass().equals(GreaterEqualsOperatorToken.class)) {
                    return ((Integer)leftLeafResult) >= ((Integer)rightLeafResult);
                } else
                if(token.getClass().equals(LessThenOperatorToken.class)) {
                    return ((Integer)leftLeafResult) > ((Integer)rightLeafResult);
                } else
                if(token.getClass().equals(LessEqualsOperatorToken.class)) {
                    return ((Integer)leftLeafResult) >= ((Integer)rightLeafResult);
                }
            } else
            if (leftType.equals(String.class) && rightType.equals(String.class)) {
                // 文字列の論理演算
                if(token.getClass().equals(EqualsOperatorToken.class)) {
                    return (leftLeafResult).equals(rightLeafResult);
                } else
                if(token.getClass().equals(NotEqualsOperatorToken.class)) {
                    return !(leftLeafResult).equals(rightLeafResult);
                }
            } else {
                if (leftType.equals(rightType)) {
                    throw new CommandRuntimeException("’" + leftType.getSimpleName() + "’ is a unknown type. " + runtime.getErrorInfo(token));
                } else {
                    throw new CommandRuntimeException("’" + leftType.getSimpleName() + "’ and '" + rightType.getSimpleName() + "' are not same type. " + runtime.getErrorInfo(token));
                }
            }
            return new Void();
    }
}
