package com.cch_lab.android.apply.prototype.tinybasic.syntax;


/**
 * 構文基底(抽象)クラス.
 */
public abstract class AbstractSyntax {
    private SyntaxPatterns syntaxPatterns;

    private SyntaxHolder prevSyntax;
    private SyntaxHolder bodySyntax;
    private SyntaxHolder nextSyntax;

    /**
     * 構文木・パターン取得.
     * @return 構文木・パターン
     */
    public abstract SyntaxPatterns getSyntaxPatterns();

    /**
     * 構文木パターン・チェック.
     * @param syntaxStream 構文・ストリーム
     * @return チェック結果 (true:構文木にマッチする、false:構文木にマッチしない)
     */
    public boolean isPattern(SyntaxStream syntaxStream) {
        if (syntaxPatterns == null) {
            syntaxPatterns = getSyntaxPatterns();
        }

        SyntaxPattern prevPattern = syntaxPatterns.getPrev();
        SyntaxPattern bodyPattern = syntaxPatterns.getBody();
        SyntaxPattern nextPattern = syntaxPatterns.getNext();

        prevSyntax = null;
        bodySyntax = null;
        nextSyntax = null;

        // パターン・チェック
        int prevIndex = syntaxStream.getScanIndex() - 1;
        int bodyIndex = syntaxStream.getScanIndex();
        int nextIndex = syntaxStream.getScanIndex() + 1;
        int prevMatchCount = 0;
        int bodyMatchCount = 0;
        int nextMatchCount = 0;

        prevMatchCount = getMatchCount(syntaxStream, prevIndex, prevPattern);
        if (!checkPattern(prevMatchCount, prevPattern)) {
            return false;
        }

        bodyMatchCount = getMatchCount(syntaxStream, bodyIndex, bodyPattern);
        if (!checkPattern(bodyMatchCount, bodyPattern)) {
            return false;
        }

        nextMatchCount = getMatchCount(syntaxStream, nextIndex, nextPattern);
        if (!checkPattern(nextMatchCount, nextPattern)) {
            return false;
        }

        // マッチングした場合の状態設定
        if (prevPattern.getPattern() == SyntaxPattern.Pattern.EXIST) {
            prevSyntax = syntaxStream.getSyntax(prevIndex, prevIndex + prevMatchCount)[0];
        }
        if (bodyPattern.getPattern() == SyntaxPattern.Pattern.EXIST) {
            bodySyntax = syntaxStream.getSyntax(bodyIndex, bodyIndex + bodyMatchCount)[0];
        }
        if (nextPattern.getPattern() == SyntaxPattern.Pattern.EXIST) {
            nextSyntax = syntaxStream.getSyntax(nextIndex, nextIndex + nextMatchCount)[0];
        }
        return true;
    }

    /**
     * 前置構文取得.
     * @return 前置構文 (isPattern（）でマッチしていなかった場合やチェック不要な場合は、ｎｕｌｌが返る)
     */
    public SyntaxHolder getPrevSyntax() {
        return prevSyntax;
    }

    /**
     * 中置構文取得.
     * @return 中置構文 (isPattern（）でマッチしていなかった場合やチェック不要な場合は、ｎｕｌｌが返る)
     */
    public SyntaxHolder getBodySyntax() {
        return bodySyntax;
    }

    /**
     * 次置構文取得.
     * @return 次置構文 (isPattern（）でマッチしていなかった場合やチェック不要な場合は、ｎｕｌｌが返る)
     */
    public SyntaxHolder getNextSyntax() {
        return nextSyntax;
    }

    /**
     * 構文パターン・マッチ回数取得.
     * @param syntaxStream  構文・ストリーム
     * @param index         構文インデックス (範囲外の場合は、チェックを行わない)
     * @param checkPattern  構文パターン
     * @return マッチ回数
     */
    int getMatchCount(SyntaxStream syntaxStream, int index, SyntaxPattern checkPattern) {
        boolean isMatch = false;
        int matchCount = 0;
        if (index < 0 || syntaxStream.size() <= index) {
            return matchCount;
        }
        SyntaxHolder syntax = syntaxStream.getSyntax(index);

        switch (checkPattern.getPattern()) {
            case NONE: {
                return matchCount;
            }
            case EXIST: {
                isMatch = isMatch(syntax, checkPattern);
                return isMatch ? ++matchCount : matchCount;
            }
            case NOT_EXIST: {
                isMatch = isMatch(syntax, checkPattern);
                return !isMatch ? ++matchCount : matchCount;
            }
            default: {
                break;
            }
        }
        return 0;
    }

    /**
     * 構文パターン・チェック.
     * @param matchCount    マッチ回数
     * @param checkPattern  構文パターン
     * @return チェック可否
     */
    boolean checkPattern(int matchCount, SyntaxPattern checkPattern) {
        switch (checkPattern.getPattern()) {
            case NONE: {
                return matchCount == 0;
            }
            case EXIST: {
                return matchCount == 1;
            }
            case NOT_EXIST: {
                return matchCount == 1;
            }
            default: {
                break;
            }
        }
        return false;
    }

    /**
     * 構文パターン・マッチチェック.
     * @param syntax        構文
     * @param checkPattern  構文パターン
     * @return マッチ可否 (true:マッチした、false:マッチしない)
     */
    boolean isMatch(SyntaxHolder syntax, SyntaxPattern checkPattern) {
        boolean isMatch = false;
        for (Class<?> checker : checkPattern.getSyntax()) {
            if (checker.equals(syntax.getTypeClass())) {
                isMatch = true;
                break;
            }
        }
        return isMatch;
    }
}
