package com.cch_lab.android.apply.prototype.tinybasic.syntax;

import java.util.ArrayList;
import java.util.List;


/**
 * 構文・マネージャ基底(抽象)クラス.
 * <p>
 *     構文の優先順位を管理するクラスです。<br/>
 *     setup メソッドで管理対象とする構文を登録してください。
 *     register メソッドで登録した順番が優先順位となります。
 * </p>
 */
public abstract class AbstractSyntaxManager {
    private int scanIndex;
    private List<Class<? extends AbstractSyntax>> syntaxList = new ArrayList<>();

    /**
     * コンストラクタ.
     */
    public AbstractSyntaxManager() {
        setup();
    }

    /**
     * セットアップ．
     * <p>
     *     構文追加(登録)設定を行ってください。
     *     <ol>登録順序(優先順序)基準
     *         <li>自己構文が、他の構文・パターンに含まれるもの(依存性が高いもの)</li>
     *         <li>優先度の高いもの順(乗除算 > 加減算)であること<li/>
     *         <li>小さな概念順(演算 < 演算式 < 文)であること</li>
     *     </ol>
     * </p>
     */
    public abstract void setup();

    /**
     * 構文登録.
     * @param syntax コマンド
     */
    public void register(Class<? extends AbstractSyntax> syntax) {
        syntaxList.add(syntax);
    }

    /**
     * 巻き戻し.
     */
    public void rewind() {
        scanIndex = 0;
    }

    /**
     * 構文保持チェック.
     * <p>
     *     次の構文があるか否かをチェックします。<br/>
     *     構文がある場合、getSyntax()で取得できます。
     * </p>
     * @return 構文保持有無 (true:あり、false:なし)
     */
    public boolean hasNext() {
        scanIndex++;
        return syntaxList.size() > scanIndex;
    }

    /**
     * 構文取得.
     * @return 構文
     */
    public AbstractSyntax getSyntax() throws InstantiationException, IllegalAccessException {
        return syntaxList.get(scanIndex).newInstance();
    }

    /**
     * 構文クラス取得.
     * @return 構文クラス
     */
    public Class<?> getSyntaxClass() {
        return syntaxList.get(scanIndex);
    }

    /**
     * 構文個数取得.
     * @return 構文個数
     */
    public int getSize() {
        return syntaxList.size();
    }

    /**
     * 構文取得.
     * @param index 構文取得位置インデックス
     * @return 構文
     */
    public AbstractSyntax getSyntax(int index) throws InstantiationException, IllegalAccessException {
        return syntaxList.get(index).newInstance();
    }

    /**
     * 構文クラス取得.
     * @param index 構文取得位置インデックス
     * @return 構文クラス
     */
    public Class<?> getSyntaxClass(int index) {
        return syntaxList.get(index);
    }

}
