package com.cch_lab.android.apply.prototype.tinybasic.syntax;

import com.cch_lab.android.apply.prototype.tinybasic.token.AbstractToken;
import com.cch_lab.android.apply.prototype.tinybasic.token.TokenStream;

import java.util.ArrayList;
import java.util.List;

/**
 * 構文・ストリーム.
 * <p>
 *     構文のストリームを表すクラスです。<br/>
 *     トークン・ストリームから構文木(構文木の入れ子構造)への変換の中間形態です。
 * </p>
 */
public class SyntaxStream {
    private List<SyntaxHolder> syntaxStream;
    private int scanIndex;

    /**
     * コンストラクタ.
     * @param tokenStream トークン・ストリーム
     */
    public SyntaxStream(TokenStream tokenStream) {
        this.syntaxStream = new ArrayList<>();
        int count = tokenStream.size();

        for (int index = 0; index < count; index++) {
            AbstractToken token = tokenStream.getToken(index);
            add(index, new SyntaxHolder(token));
        }
    }

    /**
     * 巻き戻し.
     */
    public void rewind() {
        scanIndex = 0;
    }

    /**
     * スキャン構文保持チェック.
     * @return スキャン構文保持結果 （true:スキャン対象あり、false:スキャン対象なし）
     */
    public boolean hasNext() {
        return size() > scanIndex;
    }

    /**
     * 次のスキャン構文位置に進む.
     * <ul>
     *     <li>
     *         構文がマッチしていた場合は、
     *         チェック対象の前置/中置/次置構文とチェックした構文を置き換えて
     *         入れ子にしてから、次に進みます。
     *     </li>
     *     <li>
     *         構文がマッチしていなかった場合は、そのまま次に進みます。
     *     </li>
     * </ul>
     * @param isMatch       構文マッチ可否 (true:マッチした、false:マッチしなかった)
     * @param checkedSyntax チェックした構文
     */
    public void nextSyntax(boolean isMatch, AbstractSyntax checkedSyntax) {
        if (isMatch) {
            scanIndex += replace(scanIndex, checkedSyntax);
        }
        scanIndex++;
    }

    /**
     * スキャン構文取得.
     * @return スキャン構文
     */
    public SyntaxHolder getSyntax() {
        return getSyntax(scanIndex);
    }

    /**
     * (カレント)スキャン構文位置インデックス取得.
     * @return スキャン構文位置インデックス
     */
    public int getScanIndex() {
        return scanIndex;
    }


    /**
     * 構文取得.
     * @param index 構文位置インデックス
     * @return 構文
     */
    public SyntaxHolder getSyntax(int index) {
        return syntaxStream.get(index);
    }

    /**
     * 構文配列取得.
     * @param start 構文開始位置インデックス
     * @param end   構文終了位置インデックス
     * @return 構文配列
     */
    public SyntaxHolder[] getSyntax(int start, int end) {
        int length = end - start;
        SyntaxHolder[] syntax = new SyntaxHolder[length];
        for (int index = 0; index < length; index++) {
            syntax[index] = syntaxStream.get(start + index);
        }
        return syntax;
    }
    /**
     * 構文個数取得.
     * @return 構文個数
     */
    public int size() {
        return syntaxStream.size();
    }



    /**
     * 構文木追加.
     * @param index  追加位置
     * @param syntax 追加構文
     */
    void add(int index, SyntaxHolder syntax) {
        syntaxStream.add(index, syntax);
    }

    /**
     * 構文木追加.
     * @param index  追加位置
     * @param syntax 追加構文
     */
    void add(int index, SyntaxHolder ...syntax) {
        for (int syntaxIndex = 0; syntaxIndex < syntax.length; syntaxIndex++) {
            syntaxStream.add(index + syntaxIndex, syntax[syntaxIndex]);
        }
    }

    /**
     * 構文木削除.
     * @param index 削除位置
     */
    void remove(int index) {
        syntaxStream.remove(index);
    }

    /**
     * 構文木削除.
     * @param start 削除開始位置
     * @param end   削除終了位置
     */
    void remove(int start, int end) {
        for (int index = start; index <= end; index++) {
            syntaxStream.remove(index);
        }
    }

    /**
     * 構文木置換.
     * @param index  置換位置
     * @param syntax 置換構文
     * @return 相対位置 (置換後の元位置からの相対位置を返します)
     */
    int replace(int index, AbstractSyntax syntax) {
        syntaxStream.set(index, new SyntaxHolder(syntax));

        SyntaxHolder prev = syntax.getPrevSyntax();
        SyntaxHolder body = syntax.getBodySyntax();
        SyntaxHolder next = syntax.getNextSyntax();

        int relativePosition = 0;
        if (body != null) {
            // 置換済み
        }
        if (next != null) {
            remove(index + 1);
        }
        if (prev != null) {
            remove(index - 1);
            relativePosition--;
        }
        return relativePosition;
    }

}
