package com.cch_lab.android.apply.prototype.tinybasic.syntax;

import com.cch_lab.android.apply.prototype.tinybasic.token.AbstractToken;

import java.util.ArrayList;
import java.util.List;


/**
 * 構文木・ストリーム.
 * <p>
 *     構文木のストリームを表すクラスです。
 * </p>
 */
public class SyntaxTreeStream {
    private List<SyntaxTree> syntaxTreeStream;

    /**
     * コンストラクタ.
     * @param syntaxTree 構文木
     */
    public SyntaxTreeStream(SyntaxTree syntaxTree) {
        syntaxTreeStream = extract(syntaxTree);
    }

    /**
     * 構文木取得.
     * @param index 取得位置インデックス
     * @return 構文木
     */
    public SyntaxTree getSyntaxTree(int index) {
        return syntaxTreeStream.get(index);
    }

    /**
     * 構文木位置インデックス取得.
     * @param syntaxTree 構文木
     * @return 構文木位置インデックス
     */
    public int getIndex(SyntaxTree syntaxTree) {
        int index = syntaxTreeStream.indexOf(syntaxTree);
        if (syntaxTree == null) return index;

        if (index == -1) {
            // リーフのSyntaxTreeは、syntaxを元に生成されるため中身のチェックを行う
            for (int i = 0; i < syntaxTreeStream.size(); i++) {
                if (syntaxTree.getNodeSyntax()
                        .equals(syntaxTreeStream.get(i).getNodeSyntax())) {
                    index = i;
                    break;
                }
            }
        }
        return index;
    }

    /**
     * (親)構文木位置インデックス取得.
     * <p>
     *     引数構文木を(親)に持つ構文木の位置インデックスを取得します。
     * </p>
     * @param childSyntaxTree (子)構文木
     * @return (親)構文木位置インデックス
     */
    public int getParentIndex(SyntaxTree childSyntaxTree) {
        int index = -1;
        if (childSyntaxTree == null) return index;

        // SyntaxTreeStreamは、ノードを管理するのでリーフ中身のチェックを行う
        for (int i = 0; i < syntaxTreeStream.size(); i++) {
            SyntaxTree node = syntaxTreeStream.get(i);
            if (node.getLeftLeaf() != null
                    && childSyntaxTree.getNodeSyntax()
                    .equals(node.getLeftLeaf().getNodeSyntax())) {
                index = i;
                break;
            } else
            if (node.getMiddleLeaf() != null
                    && childSyntaxTree.getNodeSyntax()
                    .equals(node.getMiddleLeaf().getNodeSyntax())) {
                index = i;
                break;
            } else
            if (node.getRightLeaf() != null
                    && childSyntaxTree.getNodeSyntax()
                    .equals(node.getRightLeaf().getNodeSyntax())) {
                index = i;
                break;
            }
        }
        return index;
    }

    /**
     * 構文木個数取得.
     * @return 構文木個数
     */
    public int size() {
        return syntaxTreeStream.size();
    }


    /**
     * （構文木展開/入れ子構造順）構文木ストリーム取得.
     * @param syntaxTree 構文木 (解析対象の構文木)
     * @return 構文木・ストリーム
     */
    private List<SyntaxTree> extract(SyntaxTree syntaxTree){
        List<SyntaxTree> syntaxTreeStream = new ArrayList<>();
        extract(syntaxTree, syntaxTreeStream);
        return syntaxTreeStream;
    }

    /**
     * （構文木展開）構文木ストリーム取得.
     * @param syntaxTree       構文木 (解析対象の構文木)
     * @param syntaxTreeStream 構文木・ストリーム (解析結果が格納される)
     */
    private void extract(SyntaxTree syntaxTree, List<SyntaxTree> syntaxTreeStream) {
        SyntaxTree leftLeaf   = syntaxTree.getLeftLeaf();
        SyntaxTree middleLeaf = syntaxTree.getMiddleLeaf();
        SyntaxTree rightLeaf  = syntaxTree.getRightLeaf();

        if (leftLeaf != null && leftLeaf.getNodeSyntax() != null) {
            syntaxTreeStream.add(leftLeaf);
            extract(leftLeaf, syntaxTreeStream);
        }
        if (middleLeaf != null && middleLeaf.getNodeSyntax() != null) {
            syntaxTreeStream.add(middleLeaf);
            extract(middleLeaf, syntaxTreeStream);
        }
        if (rightLeaf != null && rightLeaf.getNodeSyntax() != null) {
            syntaxTreeStream.add(rightLeaf);
            extract(rightLeaf, syntaxTreeStream);
        }
    }

    //＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
    // ユーティリティ・メソッド
    //＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
    /**
     * （構文木展開/定義順）構文ストリーム取得.
     * @param syntaxTree 構文木 (解析対象の構文木)
     * @return 構文・ストリーム
     */
    public static List<AbstractSyntax> extractSyntaxStream(final SyntaxTree syntaxTree) {
        List<AbstractSyntax> syntaxStream = new ArrayList<>();
        List<AbstractToken>  tokenStream  = new ArrayList<>();
        extract(syntaxTree, syntaxStream, tokenStream);
        return syntaxStream;
    }

    /**
     * （構文木展開/定義順）トークン・ストリーム取得.
     * @param syntaxTree 構文木 (解析対象の構文木)
     * @return トークン・ストリーム
     */
    public static List<AbstractToken> extractTokenStream(SyntaxTree syntaxTree) {
        List<AbstractSyntax> syntaxStream = new ArrayList<>();
        List<AbstractToken>  tokenStream  = new ArrayList<>();
        extract(syntaxTree, syntaxStream, tokenStream);
        return tokenStream;
    }

    /**
     * 構文木展開.
     * @param syntaxTree   構文木            (解析対象の構文木)
     * @param syntaxStream 構文・ストリーム    (解析結果が格納される)
     * @param tokenStream  トークン・ストリーム (解析結果が格納される)
     */
    private static void extract(SyntaxTree syntaxTree,
                                List<AbstractSyntax> syntaxStream,
                                List<AbstractToken> tokenStream) {
        SyntaxTree leftLeaf   = syntaxTree.getLeftLeaf();
        SyntaxTree middleLeaf = syntaxTree.getMiddleLeaf();
        SyntaxTree rightLeaf  = syntaxTree.getRightLeaf();

        if (leftLeaf != null) {
            extract(leftLeaf, syntaxStream, tokenStream);
            syntaxStream.add(leftLeaf.getNodeSyntax());
        } else
        if (syntaxTree.getLeftToken() != null) {
            tokenStream.add(syntaxTree.getLeftToken());
        }

        if (middleLeaf != null) {
            extract(middleLeaf, syntaxStream, tokenStream);
            syntaxStream.add(middleLeaf.getNodeSyntax());
        } else
        if (syntaxTree.getMiddleToken() != null) {
            tokenStream.add(syntaxTree.getMiddleToken());
        }

        if (rightLeaf != null) {
            extract(rightLeaf, syntaxStream, tokenStream);
            syntaxStream.add(rightLeaf.getNodeSyntax());
        } else
        if (syntaxTree.getRightToken() != null) {
            tokenStream.add(syntaxTree.getRightToken());
        }
    }

    /**
     * 構文木・文字列情報取得.
     * @param node 構文木
     * @return 構文木情報文字列
     */
    public String getSyntaxTreeInfo(SyntaxTree node) {
        SyntaxTree prev = node.getLeftLeaf();
        SyntaxTree body = node.getMiddleLeaf();
        SyntaxTree next = node.getRightLeaf();

        String nodeInfo ="node=[" + getIndex(node) + "]"+node.getNodeSyntax().getClass().getSimpleName()+"\n";
        String parentInfo = "    parent    =[" + getParentIndex(node) + "]"+"\n";
        String prevInfo = prev == null
                ? (node.getLeftToken() != null ? node.getLeftToken().getTokenSource().getSource() : null)
                : prev.getNodeSyntax() != null
                ? "["+getIndex(prev)+"]"+ prev.getNodeSyntax().getClass().getSimpleName()
                : null;
        String bodyInfo = body == null
                ? (node.getMiddleToken() != null ? node.getMiddleToken().getTokenSource().getSource() : null)
                : body.getNodeSyntax() != null
                ? "["+getIndex(body)+"]"+ body.getNodeSyntax().getClass().getSimpleName()
                : null;
        String nextInfo = next == null
                ? (node.getRightToken() != null ? node.getRightToken().getTokenSource().getSource() : null)
                : next.getNodeSyntax() != null
                ? "["+getIndex(next)+"]"+ next.getNodeSyntax().getClass().getSimpleName()
                : null;

        StringBuilder sb = new StringBuilder();
        sb.append(nodeInfo);
        sb.append(parentInfo);
        if (prevInfo != null) sb.append("    leftLeaf  =").append(prevInfo).append("\n");
        if (bodyInfo != null) sb.append("    middleLeaf=").append(bodyInfo).append("\n");
        if (nextInfo != null) sb.append("    rightLeaf =").append(nextInfo).append("\n");
        return sb.toString();
    }
}
