package com.cch_lab.android.apply.prototype.tinybasic.token;

import com.cch_lab.android.apply.prototype.tinybasic.source.SourceStream;

import java.util.regex.Matcher;
import java.lang.String;

/**
 * トークン基底(抽象)クラス.
 */
public abstract class AbstractToken {

    private TokenPattern tokenPattern;
    private TokenSource tokenSource;


    /**
     * トークン・パターン取得.
     * @return トークン・パターン
     */
    public abstract TokenPattern getTokenPattern();

    /**
     * トークンパターン・チェック.
     * @param prevToken    前置トークン (ｎｕｌｌ 可)
     * @param sourceStream ソース・ストリーム
     * @return チェック結果 (true:トークンにマッチする、false:トークンにマッチしない)
     */
    public boolean isPattern(AbstractToken prevToken, SourceStream sourceStream) {
        if (tokenPattern == null) {
            tokenPattern = getTokenPattern();
        }

        tokenSource = null;
        boolean isMatch = checkPrevPattern(prevToken);
        isMatch = tokenPattern.isNegativeCase() ? !isMatch : isMatch;
        isMatch = isMatch && checkBodyPattern(sourceStream);
        if (isMatch) {
            tokenSource = createTokenSource(sourceStream);
        }
        return isMatch;
    }

    /**
     * トークン・ソース取得.
     * @return トークン・ソース (isPattern（）でマッチしていなかった場合は、ｎｕｌｌが返る)
     */
    public TokenSource getTokenSource() {
        return tokenSource;
    }

    /**
     * 前置トークンパターン・チェック.
     * @param prevToken 前置トークン (ｎｕｌｌ 可)
     * @return チェック結果
     */
    boolean checkPrevPattern(AbstractToken prevToken) {
        Class<?>[] tokens = tokenPattern.getPrev();
        if (tokens.length != 0) {
            for (Class<?> tokenClass : tokens) {
                if (prevToken != null && prevToken.getClass().equals(tokenClass)) {
                    return true;
                }
            }
            return false;
        }
        return true;
    }

    /**
     * 自己トークンパターン・チェック.
     * @param sourceStream ソース・ストリーム
     * @return チェック結果
     */
    boolean checkBodyPattern(SourceStream sourceStream) {
        String scanSource = sourceStream.getScanSource();
        Matcher matcher = tokenPattern.getBody().matcher(scanSource.trim());
        return matcher.lookingAt();
    }

    /**
     * トークン・ソース生成.
     * @param sourceStream ソース・ストリーム
     * @return トークン・ソース (マッチしない場合は、ｎｕｌｌが返る)
     */
    TokenSource createTokenSource(SourceStream sourceStream) {
        String scanSource = sourceStream.getScanSource();
        Matcher matcher = tokenPattern.getBody().matcher(scanSource.trim());
        if (matcher.lookingAt()) {
            return new TokenSource(
                            sourceStream,
                            sourceStream.getScanRow(),
                            sourceStream.getScanColmn() + scanSource.indexOf(matcher.group()) + matcher.start(),
                            sourceStream.getScanColmn() + scanSource.indexOf(matcher.group()) + matcher.end());
        }
        return null;
    }
}
