package com.cch_lab.android.apply.prototype.tinybasic.token;

import java.util.ArrayList;
import java.util.List;

/**
 * トークン・マネージャ基底(抽象)クラス.
 * <p>
 *     トークンの優先順位を管理するクラスです。<br/>
 *     setup メソッドで管理対象とするトークンを登録してください。
 *     register メソッドで登録した順番が優先順位となります。
 * </p>
 */
public abstract class AbstractTokenManager {
    private int scanIndex;
    private List<Class<? extends AbstractToken>> tokens = new ArrayList<>();

    /**
     * コンストラクタ.
     */
    public AbstractTokenManager() {
        setup();
    }

    /**
     * セットアップ．
     * <p>
     *     トークン追加(登録)設定を行ってください。
     *     <ol>登録順序(優先順序)基準
     *         <li>自己トークンパターンに、前置トークンを必要とするもの</li>
     *         <li>自己トークン・パターンが、他のトークンに含まれるもの<br/>
     *         <i>例) 大なり比較'>'と大なりイコール比較'>='なら、大なりイコールを先にする。</i>
     *         </li>
     *     </ol>
     * </p>
     */
    public abstract void setup();

    /**
     * トークン登録.
     * @param token トークン
     */
    public void register(Class<? extends AbstractToken> token) {
        tokens.add(token);
    }

    /**
     * 巻き戻し.
     */
    public void rewind() {
        scanIndex = 0;
    }

    /**
     * トークン保持チェック.
     * <p>
     *     次のトークンがあるか否かをチェックします。<br/>
     *     トークンがある場合、getToken()で取得できます。
     * </p>
     * @return トークン保持有無 (true:あり、false:なし)
     */
    public boolean hasNext() {
        scanIndex++;
        return tokens.size() > scanIndex;
    }

    /**
     * トークン取得.
     * @return トークン
     */
    public AbstractToken getToken() throws InstantiationException, IllegalAccessException {
        return tokens.get(scanIndex).newInstance();
    }

    /**
     * トークン個数取得.
     * @return トークン個数
     */
    public int getSize() {
        return tokens.size();
    }

    /**
     * トークン取得.
     * @param index トークン取得位置インデックス
     * @return トークン
     */
    public AbstractToken getToken(int index) throws InstantiationException, IllegalAccessException {
        return tokens.get(index).newInstance();
    }
}
